import pytest

import titan.resources as res

from titan.enums import ResourceType
from titan.resources.resource import ResourcePointer


def test_fqn_construction():
    ptr = ResourcePointer(name="my_db.my_schema.my_table", resource_type=ResourceType.TABLE)
    assert ptr.name == "my_table"
    assert ptr.container.name == "my_schema"
    assert ptr.container.container.name == "my_db"


def test_resource_pointer_type_checking():
    invalid_pointer = ResourcePointer(name="my_network_rule", resource_type=ResourceType.DATABASE)
    with pytest.raises(TypeError):
        res.NetworkPolicy(name="my_network_policy", allowed_network_rule_list=[invalid_pointer])

    with pytest.raises(TypeError):
        res.NetworkPolicy(name="my_network_policy", allowed_network_rule_list=[111])



from dataclasses import dataclass

from ..enums import EncryptionType, ParseableEnum, ResourceType
from ..props import (
    BoolProp,
    EnumProp,
    PropList,
    Props,
    PropSet,
    StringProp,
    StructProp,
)
from ..resource_name import ResourceName
from ..scope import AccountScope, AnonymousScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role


class ExternalVolumeStorageProvider(ParseableEnum):
    AZURE = "AZURE"
    GCS = "GCS"
    S3 = "S3"
    S3GOV = "S3GOV"


@dataclass(unsafe_hash=True)
class _ExternalVolumeStorageLocation(ResourceSpec):
    name: str
    storage_provider: ExternalVolumeStorageProvider
    storage_base_url: str
    encryption: dict = None
    storage_aws_role_arn: str = None
    storage_aws_external_id: str = None
    # storage_allowed_locations: list[str] = field(default=None, metadata={"known_after_apply": True})
    # storage_aws_iam_user_arn: str = field(default=None, metadata={"known_after_apply": True})

    def __post_init__(self):
        super().__post_init__()
        if self.encryption is None:
            self.encryption = {"type": EncryptionType.NONE}


class ExternalVolumeStorageLocation(Resource):
    resource_type = ResourceType.EXTERNAL_VOLUME_STORAGE_LOCATION
    props = Props(
        name=StringProp("name"),
        storage_provider=EnumProp("storage_provider", ExternalVolumeStorageProvider, quoted=True),
        storage_base_url=StringProp("storage_base_url"),
        encryption=PropSet(
            "encryption",
            Props(
                type=EnumProp(
                    "type",
                    [
                        EncryptionType.AWS_SSE_S3,
                        EncryptionType.AWS_SSE_KMS,
                        EncryptionType.GCS_SSE_KMS,
                        EncryptionType.NONE,
                    ],
                    quoted=True,
                ),
                kms_key_id=StringProp("kms_key_id"),
            ),
        ),
        storage_aws_role_arn=StringProp("storage_aws_role_arn"),
        storage_aws_external_id=StringProp("storage_aws_external_id"),
    )
    scope = AnonymousScope()
    spec = _ExternalVolumeStorageLocation
    serialize_inline = True

    def __init__(
        self,
        name: str,
        storage_provider: ExternalVolumeStorageProvider,
        storage_base_url: str,
        encryption: dict = None,
        storage_aws_role_arn: str = None,
        storage_aws_external_id: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._data: _ExternalVolumeStorageLocation = _ExternalVolumeStorageLocation(
            name=name,
            storage_provider=storage_provider,
            storage_base_url=storage_base_url,
            encryption=encryption,
            storage_aws_role_arn=storage_aws_role_arn,
            storage_aws_external_id=storage_aws_external_id,
        )


@dataclass(unsafe_hash=True)
class _ExternalVolume(ResourceSpec):
    name: ResourceName
    owner: Role = "ACCOUNTADMIN"
    storage_locations: list[ExternalVolumeStorageLocation] = None
    allow_writes: bool = True
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.storage_locations is None or len(self.storage_locations) == 0:
            raise ValueError("storage_locations is required")


class ExternalVolume(NamedResource, Resource):
    resource_type = ResourceType.EXTERNAL_VOLUME
    props = Props(
        storage_locations=PropList(
            "storage_locations",
            StructProp(ExternalVolumeStorageLocation.props),
        ),
        allow_writes=BoolProp("allow_writes"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _ExternalVolume

    def __init__(
        self,
        name: str,
        owner: str = "ACCOUNTADMIN",
        storage_locations: list[dict] = None,
        allow_writes: bool = True,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ExternalVolume = _ExternalVolume(
            name=self._name,
            owner=owner,
            storage_locations=storage_locations,
            allow_writes=allow_writes,
            comment=comment,
        )



from typing import Optional, Union

from .enums import DataType


def convert_to_canonical_data_type(data_type: Union[str, DataType, None]) -> Optional[str]:
    if data_type is None:
        return None
    if isinstance(data_type, DataType):
        data_type = str(data_type)
    data_type = data_type.upper()
    if data_type in (
        "NUMBER",
        "DECIMAL",
        "DEC",
        "NUMERIC",
        "INT",
        "INTEGER",
        "BIGINT",
        "SMALLINT",
        "TINYINT",
        "BYTEINT",
    ):
        return "NUMBER(38,0)"
    if data_type in (
        "FLOAT",
        "FLOAT4",
        "FLOAT8",
        "REAL",
        "DOUBLE",
        "DOUBLE PRECISION",
        "REAL",
    ):
        return "FLOAT"
    if data_type in ("BOOLEAN", "BOOL"):
        return "BOOLEAN"
    if data_type in (
        "VARCHAR",
        "STRING",
        "TEXT",
        "NVARCHAR",
        "NVARCHAR2",
        "CHAR VARYING",
        "NCHAR VARYING",
    ):
        return "VARCHAR(16777216)"
    if data_type in ("CHAR", "CHARACTER", "NCHAR"):
        return "VARCHAR(1)"
    if data_type in ("BINARY", "VARBINARY"):
        return "BINARY(8388608)"
    if data_type in ("DATETIME", "TIMESTAMP", "TIMESTAMP_NTZ"):
        return "TIMESTAMP_NTZ(9)"
    if data_type in ("TIMESTAMP_LTZ", "TIMESTAMP_TZ"):
        return "TIMESTAMP_LTZ(9)"
    if data_type in ("TIME"):
        return "TIME(9)"
    return data_type



import pytest

from titan import resources as res
from titan.blueprint import (
    Blueprint,
    CreateResource,
    MissingPrivilegeException,
    TransferOwnership,
    UpdateResource,
    compile_plan_to_sql,
)
from titan.enums import AccountEdition
from titan.identifiers import parse_URN
from titan.resource_name import ResourceName


@pytest.fixture
def session_ctx() -> dict:
    return {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "role": "SYSADMIN",
        "available_roles": [
            "SYSADMIN",
            "USERADMIN",
            "ACCOUNTADMIN",
            "SECURITYADMIN",
            "PUBLIC",
        ],
    }


@pytest.fixture
def remote_state() -> dict:
    return {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
    }


def test_default_owner(session_ctx, remote_state):
    warehouse = res.Warehouse(name="test_warehouse")
    assert warehouse._data.owner.name == "SYSADMIN"
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].after["owner"] == "SYSADMIN"


def test_non_default_owner(session_ctx, remote_state):
    warehouse = res.Warehouse(name="test_warehouse", owner="ACCOUNTADMIN")
    assert warehouse._data.owner.name == "ACCOUNTADMIN"
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].after["owner"] == "ACCOUNTADMIN"


def test_custom_role_owner(session_ctx, remote_state):
    role = res.Role(name="CUSTOMROLE")
    grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    warehouse = res.Warehouse(name="test_warehouse", owner=role)
    assert warehouse._data.owner.name == "CUSTOMROLE"
    blueprint = Blueprint(resources=[role, grant, warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 3
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:role/CUSTOMROLE")
    assert isinstance(plan[1], CreateResource)
    assert plan[1].urn == parse_URN("urn::ABCD123:role_grant/CUSTOMROLE?role=SYSADMIN")
    assert isinstance(plan[2], CreateResource)
    assert plan[2].urn == parse_URN("urn::ABCD123:warehouse/test_warehouse")
    assert plan[2].after["owner"] == "CUSTOMROLE"


# def test_invalid_custom_role_owner(session_ctx):
#     role = res.Role(name="INVALIDROLE")
#     warehouse = res.Warehouse(name="test_warehouse", owner=role)
#     blueprint = Blueprint(resources=[role, warehouse])
#     with pytest.raises(InvalidOwnerException):
#         blueprint.generate_manifest(session_ctx)


def test_transfer_ownership(session_ctx, remote_state):
    remote_state = remote_state.copy()
    remote_state[parse_URN("urn::ABCD123:role/test_role")] = {
        "name": str(ResourceName("test_role")),
        "owner": "ACCOUNTADMIN",
        "comment": None,
    }

    role = res.Role(name="test_role", owner="USERADMIN")
    blueprint = Blueprint(resources=[role])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], TransferOwnership)
    assert plan[0].from_owner == "ACCOUNTADMIN"
    assert plan[0].to_owner == "USERADMIN"
    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE ACCOUNTADMIN"
    assert sql_commands[2] == "GRANT OWNERSHIP ON ROLE TEST_ROLE TO ROLE USERADMIN COPY CURRENT GRANTS"


def test_transfer_ownership_with_changes(session_ctx, remote_state):
    remote_state = remote_state.copy()
    remote_state[parse_URN("urn::ABCD123:role/test_role")] = {
        "name": str(ResourceName("test_role")),
        "owner": "ACCOUNTADMIN",
        "comment": None,
    }

    role = res.Role(name="test_role", comment="This comment has been added", owner="USERADMIN")
    blueprint = Blueprint(resources=[role])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 2
    assert isinstance(plan[0], UpdateResource)
    assert plan[0].after["comment"] == "This comment has been added"
    assert isinstance(plan[1], TransferOwnership)
    assert plan[1].from_owner == "ACCOUNTADMIN"
    assert plan[1].to_owner == "USERADMIN"
    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE ACCOUNTADMIN"
    assert sql_commands[2] == "ALTER ROLE TEST_ROLE SET COMMENT = $$This comment has been added$$"
    assert sql_commands[3] == "USE ROLE ACCOUNTADMIN"
    assert sql_commands[4] == "GRANT OWNERSHIP ON ROLE TEST_ROLE TO ROLE USERADMIN COPY CURRENT GRANTS"


def test_resource_is_transferred_to_custom_role_owner(session_ctx, remote_state):
    session_ctx = session_ctx.copy()
    session_ctx["available_roles"].append(ResourceName("test_role"))

    warehouse = res.Warehouse(name="test_warehouse", owner="test_role")
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:warehouse/test_warehouse")

    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE SYSADMIN"
    assert sql_commands[2].startswith("CREATE WAREHOUSE TEST_WAREHOUSE")
    assert sql_commands[3] == "GRANT OWNERSHIP ON WAREHOUSE TEST_WAREHOUSE TO ROLE TEST_ROLE COPY CURRENT GRANTS"


def test_resource_cant_be_created(remote_state):
    session_ctx = {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "role": "TEST_ROLE",
        "available_roles": [
            "TEST_ROLE",
        ],
    }
    warehouse = res.Warehouse(name="test_warehouse", owner="test_role")
    blueprint = Blueprint(resources=[warehouse])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:warehouse/test_warehouse")

    with pytest.raises(MissingPrivilegeException):
        compile_plan_to_sql(session_ctx, plan)


def test_grant_with_grant_admin_custom_role(remote_state):
    session_ctx = {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "role": "GRANT_ADMIN",
        "available_roles": [
            "GRANT_ADMIN",
        ],
    }

    grant = res.RoleGrant(role="GRANT_ADMIN", to_role="SYSADMIN")
    blueprint = Blueprint(resources=[grant])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:role_grant/GRANT_ADMIN?role=SYSADMIN")
    compile_plan_to_sql(session_ctx, plan)


def test_tag_reference_with_tag_admin_custom_role():
    session_ctx = {
        "account": "SOMEACCT",
        "account_edition": AccountEdition.ENTERPRISE,
        "account_locator": "ABCD123",
        "role": "TAG_ADMIN",
        "available_roles": [
            "TAG_ADMIN",
        ],
        "tags": ["tags.tags.cost_center"],
    }

    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
    }

    tag_reference = res.TagReference(
        object_name="SOME_ROLE",
        object_domain="ROLE",
        tags={"tags.tags.cost_center": "finance"},
    )
    blueprint = Blueprint(resources=[tag_reference])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 1
    assert isinstance(plan[0], CreateResource)
    assert plan[0].urn == parse_URN("urn::ABCD123:tag_reference/SOME_ROLE?domain=ROLE")
    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert len(sql_commands) == 3
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE TAG_ADMIN"
    assert sql_commands[2] == "ALTER ROLE SOME_ROLE SET TAG tags.tags.cost_center='finance'"


def test_owner_is_database_role(session_ctx):
    remote_state = {
        parse_URN("urn::ABCD123:account/ACCOUNT"): {},
        parse_URN("urn::ABCD123:database/SOME_DATABASE"): {"owner": "SYSADMIN"},
        parse_URN("urn::ABCD123:schema/SOME_DATABASE.PUBLIC"): {"owner": "SYSADMIN"},
    }

    def _plan_for_resources(resources):
        blueprint = Blueprint(resources=resources)
        manifest = blueprint.generate_manifest(session_ctx)
        return blueprint._plan(remote_state, manifest)

    # Specify owner as a string
    database_role = res.DatabaseRole(
        name="SOME_DATABASE_ROLE",
        database="SOME_DATABASE",
    )
    schema = res.Schema(
        name="SOME_SCHEMA",
        database="SOME_DATABASE",
        owner="SOME_DATABASE.SOME_DATABASE_ROLE",
    )
    plan = _plan_for_resources([database_role, schema])
    assert len(plan) == 2

    # Specify owner as a resource
    database_role = res.DatabaseRole(
        name="SOME_DATABASE_ROLE",
        database="SOME_DATABASE",
    )
    schema = res.Schema(
        name="SOME_SCHEMA",
        database="SOME_DATABASE",
        owner=database_role,
    )
    plan = _plan_for_resources([database_role, schema])
    assert len(plan) == 2


def test_blueprint_create_resource_with_database_role_owner(session_ctx, remote_state):

    database = res.Database(name="SOME_DATABASE")
    database_role = res.DatabaseRole(
        name="SOME_DATABASE_ROLE",
        database=database,
        owner="SYSADMIN",
    )
    schema = res.Schema(name="test_schema", database=database, owner=database_role)
    blueprint = Blueprint(resources=[database, database_role, schema])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 3
    sql_commands = compile_plan_to_sql(session_ctx, plan)
    # assert len(sql_commands) == 8
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE SYSADMIN"
    assert sql_commands[2].startswith("CREATE DATABASE SOME_DATABASE")
    assert sql_commands[3] == "USE ROLE SYSADMIN"
    assert sql_commands[4] == "CREATE DATABASE ROLE SOME_DATABASE.SOME_DATABASE_ROLE"
    assert sql_commands[5] == "USE ROLE SYSADMIN"
    assert sql_commands[6].startswith("CREATE SCHEMA SOME_DATABASE.TEST_SCHEMA")
    assert (
        sql_commands[7]
        == "GRANT OWNERSHIP ON SCHEMA SOME_DATABASE.TEST_SCHEMA TO DATABASE ROLE SOME_DATABASE.SOME_DATABASE_ROLE COPY CURRENT GRANTS"
    )


def test_database_with_custom_owner_modifies_public_schema_owner(session_ctx, remote_state):
    role = res.Role(name="CUSTOM_ROLE")
    role_grant = res.RoleGrant(role=role, to_role="SYSADMIN")
    database = res.Database(name="SOME_DATABASE", owner=role)
    blueprint = Blueprint(resources=[database, role, role_grant])
    manifest = blueprint.generate_manifest(session_ctx)
    plan = blueprint._plan(remote_state, manifest)
    assert len(plan) == 3
    sql_commands = compile_plan_to_sql(session_ctx, plan)
    assert len(sql_commands) == 9
    assert sql_commands[0] == "USE SECONDARY ROLES ALL"
    assert sql_commands[1] == "USE ROLE USERADMIN"
    assert sql_commands[2] == "CREATE ROLE CUSTOM_ROLE"
    assert sql_commands[3] == "USE ROLE SECURITYADMIN"
    assert sql_commands[4] == "GRANT ROLE CUSTOM_ROLE TO ROLE SYSADMIN"
    assert sql_commands[5] == "USE ROLE SYSADMIN"
    assert sql_commands[6].startswith("CREATE DATABASE SOME_DATABASE")
    assert sql_commands[7] == "GRANT OWNERSHIP ON DATABASE SOME_DATABASE TO ROLE CUSTOM_ROLE COPY CURRENT GRANTS"
    assert sql_commands[8] == "GRANT OWNERSHIP ON SCHEMA SOME_DATABASE.PUBLIC TO ROLE CUSTOM_ROLE COPY CURRENT GRANTS"



from titan.parse import parse_region


def test_parse_region():
    assert parse_region("AWS_US_WEST_2") == {"cloud": "AWS", "cloud_region": "US_WEST_2"}
    assert parse_region("PUBLIC.AWS_US_WEST_2") == {
        "region_group": "PUBLIC",
        "cloud": "AWS",
        "cloud_region": "US_WEST_2",
    }
    assert parse_region("AZURE_WESTUS2") == {"cloud": "AZURE", "cloud_region": "WESTUS2"}
    assert parse_region("GCP_EUROPE_WEST4") == {"cloud": "GCP", "cloud_region": "EUROPE_WEST4"}

    assert parse_region("AWS_US_GOV_WEST_1_FHPLUS") == {"cloud": "AWS", "cloud_region": "US_GOV_WEST_1_FHPLUS"}
    assert parse_region("AWS_US_GOV_WEST_1_DOD") == {"cloud": "AWS", "cloud_region": "US_GOV_WEST_1_DOD"}
    assert parse_region("AWS_AP_SOUTHEAST_1") == {"cloud": "AWS", "cloud_region": "AP_SOUTHEAST_1"}
    assert parse_region("AWS_EU_CENTRAL_1") == {"cloud": "AWS", "cloud_region": "EU_CENTRAL_1"}

    assert parse_region("AZURE_CANADACENTRAL") == {"cloud": "AZURE", "cloud_region": "CANADACENTRAL"}
    assert parse_region("AZURE_NORTHEUROPE") == {"cloud": "AZURE", "cloud_region": "NORTHEUROPE"}
    assert parse_region("AZURE_SWITZERLANDNORTH") == {"cloud": "AZURE", "cloud_region": "SWITZERLANDNORTH"}
    assert parse_region("AZURE_USGOVVIRGINIA") == {"cloud": "AZURE", "cloud_region": "USGOVVIRGINIA"}

    assert parse_region("GCP_US_CENTRAL1") == {"cloud": "GCP", "cloud_region": "US_CENTRAL1"}
    assert parse_region("GCP_EUROPE_WEST2") == {"cloud": "GCP", "cloud_region": "EUROPE_WEST2"}
    assert parse_region("GCP_EUROPE_WEST3") == {"cloud": "GCP", "cloud_region": "EUROPE_WEST3"}

    assert parse_region("PUBLIC.AWS_EU_CENTRAL_1") == {
        "region_group": "PUBLIC",
        "cloud": "AWS",
        "cloud_region": "EU_CENTRAL_1",
    }
    assert parse_region("PUBLIC.AZURE_WESTEUROPE") == {
        "region_group": "PUBLIC",
        "cloud": "AZURE",
        "cloud_region": "WESTEUROPE",
    }
    assert parse_region("PUBLIC.GCP_US_CENTRAL1") == {
        "region_group": "PUBLIC",
        "cloud": "GCP",
        "cloud_region": "US_CENTRAL1",
    }

    assert parse_region("SOME_VALUE.GCP_US_CENTRAL1") == {
        "region_group": "SOME_VALUE",
        "cloud": "GCP",
        "cloud_region": "US_CENTRAL1",
    }



from dataclasses import dataclass
from typing import Any

from ..enums import ResourceType
from ..parse import parse_alter_account_parameter
from ..props import Props
from ..resource_name import ResourceName
from ..scope import AccountScope
from .resource import NamedResource, Resource, ResourceSpec


@dataclass(unsafe_hash=True)
class _AccountParameter(ResourceSpec):
    name: ResourceName
    value: Any


class AccountParameter(NamedResource, Resource):
    """
    Description:
        An account parameter in Snowflake that allows you to set or alter account-level parameters.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/alter-account

    Fields:
        name (string, required): The name of the account parameter.
        value (Any, required): The value to set for the account parameter.

    Python:

        ```python
        account_parameter = AccountParameter(
            name="some_parameter",
            value="some_value",
        )
        ```

    Yaml:

        ```yaml
        account_parameters:
          - name: some_parameter
            value: some_value
        ```
    """

    resource_type = ResourceType.ACCOUNT_PARAMETER
    props = Props()
    scope = AccountScope()
    spec = _AccountParameter

    def __init__(self, name: str, value: Any, **kwargs):
        super().__init__(name=name, **kwargs)
        self._data: _AccountParameter = _AccountParameter(
            name=self._name,
            value=value,
        )

    @classmethod
    def from_sql(cls, sql):
        props = parse_alter_account_parameter(sql)
        return cls(**props)
